import 'dart:collection';
import 'dart:core';

// A sparse Fenwick tree
// See https://en.wikipedia.org/wiki/Fenwick_tree
class SparsePrefixSumList with ListMixin<num> implements List<num> {
  SparsePrefixSumList(num defaultValue, [int initialLength])
      : assert(initialLength == null || initialLength >= 0),
        _length = initialLength ?? 0,
        _defaultValue = defaultValue,
        _hasNegativeValue =
            initialLength != null && initialLength > 0 && defaultValue < 0,
        _tree = Map<int, num>(),
        _values = Map<int, num>() {
    // initialize index
    _addTree(1, 0);
  }

  @override
  num operator [](int index) {
    assert(index >= 0 && index < length);
    return _getAt(index + 1);
  }

  @override
  void operator []=(int index, num value) {
    assert(index >= 0 && index < length);
    _setAt(index + 1, value);
  }

  num getPrefixSum(int index) {
    assert(index >= 0 && index < length);
    index++;
    return _getPrefixSum(index);
  }

  /// Will return -1 if no index can be found.
  int findPrefixSumBefore(num sum) {
    assert(!_hasNegativeValue);
    return _findPrefixSumBefore(sum) - 1;
  }

  set defaultValue(num newValue) {
    _tree.clear();
    _defaultValue = newValue;
    _hasNegativeValue = _values.length < length && _defaultValue < 0;
    void updateIndex(int index, num value) {
      _hasNegativeValue = _hasNegativeValue || value < 0;
      final diff = value - _defaultValue;
      _addTree(index, diff);
    }

    _values.forEach(updateIndex);
  }

  num get defaultValue => _defaultValue;

  set length(int newLength) {
    // fixme: delete unneeded values
    if (newLength > _indexBound) {
      final prevBound = _indexBound;
      _length = newLength;
      _addTree(prevBound, 0);
    } else {
      _length = newLength;
    }
  }

  int get length {
    return _length;
  }

  num _getAt(int index) {
    return _values[index] ?? _defaultValue;
  }

  void _setAt(int index, num value) {
    if (value < 0) _hasNegativeValue = true;
    final diff = value - (_values[index] ?? _defaultValue);
    _values[index] = value;
    _addTree(index, diff);
  }

  num _getPrefixSum(int index) {
    num sum = 0;
    for (int node = index; node > 0; node = node ^ _lsb(node)) {
      sum += _getTreeAt(node);
    }
    return sum;
  }

  int _findPrefixSumBefore(num sum) {
    num partialSum = 0;
    int index = 0;
    for (int branch = _indexBound;
        branch > 0;
        branch = branch == 1 ? 0 : branch >> 1) {
      final guessIndex = index | branch;
      if (partialSum + _getTreeAt(guessIndex) > sum) continue;
      partialSum += _getTreeAt(guessIndex);
      index = guessIndex;
    }
    return index;
  }

  num _getTreeAt(int index) {
    if (_tree[index] != null) return _tree[index];
    final distanceToNearestBase = _lsb(index);
    final defaultTo = distanceToNearestBase * _defaultValue;
    return defaultTo;
  }

  int get _indexBound => _msb(length) << 1;

  void _addTree(int from, num diff) {
    for (int index = from; index <= _indexBound; index = index + _lsb(index)) {
      _tree[index] = _getTreeAt(index) + diff;
    }
  }

  int _length;
  num _defaultValue;
  final Map<int, num> _values;
  final Map<int, num> _tree;
  bool _hasNegativeValue;
}

int _msb(int number) {
  return number > 0 ? 1 << (number.bitLength - 1) : 0;
}

int _lsb(int number) {
  return number & -number;
}
